"""Interact with the local API of La Marzocco machines."""

import logging
from typing import Any, Callable

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    WSMsgType,
    ClientWSTimeout,
    ClientConnectionError,
)
from aiohttp.client_exceptions import ClientError, InvalidURL

from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.const import DEFAULT_PORT
from pylamarzocco.exceptions import AuthFail, RequestNotSuccessful
from pylamarzocco.helpers import is_success

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoLocalClient:
    """Class to interact with machine via local API."""

    def __init__(
        self,
        host: str,
        local_bearer: str,
        local_port: int = DEFAULT_PORT,
        client: ClientSession | None = None,
    ) -> None:
        self._host = host
        self._local_port = local_port
        self._local_bearer = local_bearer

        self.websocket: ClientWebSocketResponse | None = None
        self.websocket_disconnected = False

        if client is None:
            self._client = ClientSession()
        else:
            self._client = client

    async def get_config(self) -> dict[str, Any]:
        """Get current config of machine from local API."""
        return await self._get_config(
            self._client,
            self._host,
            self._local_bearer,
            self._local_port,
        )

    @staticmethod
    async def validate_connection(
        client: ClientSession,
        host: str,
        token: str,
        port: int = DEFAULT_PORT,
        cloud_details: tuple[LaMarzoccoCloudClient, str] | None = None,
    ) -> bool:
        """Validate the connection details to the local API."""
        try:
            await LaMarzoccoLocalClient._get_config(client, host, token, port)
        except AuthFail:
            # try to activate the local API
            if cloud_details is not None:
                cloud_client, serial = cloud_details
                try:
                    await cloud_client.token_command(serial)
                    await LaMarzoccoLocalClient._get_config(client, host, token, port)
                except (AuthFail, RequestNotSuccessful) as ex:
                    _LOGGER.error(ex)
                    return False
        except RequestNotSuccessful as ex:
            _LOGGER.error(ex)
            return False
        return True

    @staticmethod
    async def _get_config(
        client: ClientSession,
        host: str,
        token: str,
        port: int = DEFAULT_PORT,
    ) -> dict[str, Any]:
        """Get current config of machine from local API."""
        headers = {"Authorization": f"Bearer {token}"}

        try:
            response = await client.get(
                f"http://{host}:{port}/api/v1/config", headers=headers
            )
        except ClientError as ex:
            raise RequestNotSuccessful(
                f"Requesting local API failed with exception: {ex}"
            ) from ex
        if is_success(response):
            return await response.json()
        if response.status == 403:
            raise AuthFail("Local API returned 403.")
        raise RequestNotSuccessful(
            f"Querying local API failed with statuscode: {response.status}"
            + f"response: {await response.text()}"
        )

    async def websocket_connect(
        self,
        callback: Callable[[str | bytes], None] | None = None,
    ) -> None:
        """Connect to the websocket of the machine."""

        try:
            async with await self._client.ws_connect(
                f"ws://{self._host}:{self._local_port}/api/v1/streaming",
                headers={"Authorization": f"Bearer {self._local_bearer}"},
                timeout=ClientWSTimeout(ws_receive=25, ws_close=10.0),
            ) as ws:
                self.websocket = ws
                if self.websocket_disconnected:
                    _LOGGER.warning("Websocket reconnected")
                    self.websocket_disconnected = False
                async for msg in ws:
                    if msg.type in (WSMsgType.CLOSING, WSMsgType.CLOSED):
                        _LOGGER.warning("Websocket disconnected gracefully")
                        self.websocket_disconnected = True
                        break
                    if msg.type == WSMsgType.ERROR:
                        _LOGGER.warning(
                            "Websocket disconnected with error %s", ws.exception()
                        )
                        self.websocket_disconnected = True
                        break
                    _LOGGER.debug("Received websocket message: %s", msg)
                    if callback is not None:
                        try:
                            callback(msg.data)
                        except Exception as ex:  # pylint: disable=broad-except
                            _LOGGER.exception("Error during callback: %s", ex)
        except TimeoutError as err:
            if not self.websocket_disconnected:
                _LOGGER.warning("Websocket disconnected: Connection timed out")
                self.websocket_disconnected = True
            _LOGGER.debug("Websocket timeout: %s", err)
        except ClientConnectionError as err:
            if not self.websocket_disconnected:
                _LOGGER.warning("Websocket disconnected: Could not connect: %s", err)
                self.websocket_disconnected = True
            _LOGGER.debug("Websocket disconnected: Could not connect: %s", err)
        except InvalidURL:
            _LOGGER.error("Invalid URL for websocket.")



"""Import for ease of use."""

from pylamarzocco.clients.bluetooth import LaMarzoccoBluetoothClient
from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.clients.local import LaMarzoccoLocalClient
from pylamarzocco.devices.grinder import LaMarzoccoGrinder
from pylamarzocco.devices.machine import LaMarzoccoMachine



"""Module for La Marzocco coffee machine."""

from __future__ import annotations

import json
import logging
from collections.abc import Callable
from copy import deepcopy
from time import time
from typing import Any

from pylamarzocco.clients.bluetooth import LaMarzoccoBluetoothClient
from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.clients.local import LaMarzoccoLocalClient
from pylamarzocco.const import (
    BoilerType,
    FirmwareType,
    MachineModel,
    PhysicalKey,
    PrebrewMode,
    SmartStandbyMode,
    SteamLevel,
)
from pylamarzocco.exceptions import ClientNotInitialized, UnknownWebSocketMessage
from pylamarzocco.helpers import (
    parse_boilers,
    parse_brew_by_weight_settings,
    parse_cloud_statistics,
    parse_coffee_doses,
    parse_preinfusion_settings,
    parse_scale,
    parse_smart_standby,
    parse_wakeup_sleep_entries,
)
from pylamarzocco.devices.base import LaMarzoccoBaseDevice
from pylamarzocco.models import (
    LaMarzoccoCoffeeStatistics,
    LaMarzoccoMachineConfig,
    LaMarzoccoSmartStandby,
    LaMarzoccoWakeUpSleepEntry,
)

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoMachine(LaMarzoccoBaseDevice):
    """Class for La Marzocco coffee machine"""

    def __init__(
        self,
        model: MachineModel,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
        bluetooth_client: LaMarzoccoBluetoothClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoCoffeeMachine instance"""

        super().__init__(
            model=model,
            serial_number=serial_number,
            name=name,
            cloud_client=cloud_client,
            local_client=local_client,
            bluetooth_client=bluetooth_client,
        )
        self.config: LaMarzoccoMachineConfig = LaMarzoccoMachineConfig(
            turned_on=False,
            boilers={},
            prebrew_mode=PrebrewMode.DISABLED,
            plumbed_in=False,
            prebrew_configuration={},
            dose_hot_water=0,
            doses={},
            water_contact=False,
            brew_active=False,
            backflush_enabled=False,
            brew_active_duration=0,
            smart_standby=LaMarzoccoSmartStandby(
                enabled=False,
                minutes=10,
                mode=SmartStandbyMode.POWER_ON,
            ),
            wake_up_sleep_entries={},
        )
        self.statistics: LaMarzoccoCoffeeStatistics = LaMarzoccoCoffeeStatistics(
            drink_stats={},
            continous=0,
            total_flushes=0,
        )

        self._notify_callback: Callable[[], None] | None = None
        self._system_info: dict[str, Any] | None = None
        self.timestamp_last_websocket_msg: float | None = None

    @classmethod
    async def create(cls, *args: Any, **kwargs: Any) -> LaMarzoccoMachine:
        """Create a new LaMarzoccoMachine instance"""

        self = cls(*args, **kwargs)
        await self.get_config()
        if self._cloud_client is not None:
            await self.get_firmware()
            await self.get_statistics()
        return self

    @property
    def full_model_name(self) -> str:
        """Return the full model name"""

        if self.model == MachineModel.LINEA_MICRA:
            return "Linea Micra"

        return self.model

    @property
    def websocket_connected(self) -> bool:
        """Return the connection status of the websocket client."""
        if self._local_client and self._local_client.websocket:
            return not self._local_client.websocket.closed
        return False

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        super().parse_config(raw_config)
        self._raw_config = raw_config
        self.config.turned_on = raw_config["machineMode"] == "BrewingMode"
        self.config.plumbed_in = raw_config["isPlumbedIn"]
        self.config.water_contact = raw_config["tankStatus"]
        self.config.backflush_enabled = raw_config["isBackFlushEnabled"]
        self.config.doses, self.config.dose_hot_water = parse_coffee_doses(raw_config)
        self.config.boilers = parse_boilers(raw_config["boilers"])
        self.config.prebrew_mode, self.config.prebrew_configuration = (
            parse_preinfusion_settings(raw_config)
        )
        self.config.smart_standby = parse_smart_standby(
            raw_config.get("smartStandBy", {})
        )
        self.config.wake_up_sleep_entries = parse_wakeup_sleep_entries(
            raw_config.get("wakeUpSleepEntries", {})
        )
        self.config.scale = parse_scale(raw_config)
        self.config.bbw_settings = parse_brew_by_weight_settings(raw_config)

    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

        self.statistics = parse_cloud_statistics(raw_statistics)

    async def set_power(
        self,
        enabled: bool,
    ) -> bool:
        """Turn power of machine on or off"""

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_power",
            enabled=enabled,
        ):
            self.config.turned_on = enabled
            self.config.boilers[BoilerType.COFFEE].enabled = enabled
            return True
        return False

    async def set_steam(
        self,
        enabled: bool,
    ) -> bool:
        """Turn Steamboiler on or off"""

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_steam",
            enabled=enabled,
        ):
            self.config.boilers[BoilerType.STEAM].enabled = enabled
            return True
        return False

    async def set_temp(
        self,
        boiler: BoilerType,
        temperature: float,
    ) -> bool:
        """Set target temperature for boiler"""

        if boiler == BoilerType.STEAM:
            if self.model == MachineModel.LINEA_MICRA:
                if temperature not in SteamLevel:
                    msg = "Steam temp must be one of 126, 128, 131 (°C)"
                    _LOGGER.debug(msg)
                    raise ValueError(msg)
            elif self.model == MachineModel.LINEA_MINI:
                _LOGGER.error("Steam temp is not supported on Linea Mini.")
                return False
        else:
            if temperature > 104 or temperature < 85:
                msg = "Coffee temp must be between 85 and 104 (°C)"
                _LOGGER.debug(msg)
                raise ValueError(msg)

            temperature = round(temperature, 1)

        if await self._bluetooth_command_with_cloud_fallback(
            command="set_temp",
            boiler=boiler,
            temperature=temperature,
        ):

            self.config.boilers[boiler].target_temperature = temperature
            return True
        return False

    async def set_steam_level(
        self,
        level: SteamLevel,
    ) -> bool:
        """Set steam level"""
        return await self.set_temp(BoilerType.STEAM, level)

    async def set_prebrew_mode(self, mode: PrebrewMode) -> bool:
        """Set preinfusion mode"""

        if mode == PrebrewMode.PREINFUSION and not self.config.plumbed_in:
            msg = "Pre-Infusion can only be enabled when plumbin is enabled."
            _LOGGER.debug(msg)
            raise ValueError(msg)

        if await self.cloud_client.set_prebrew_mode(self.serial_number, mode):
            self.config.prebrew_mode = mode
            return True
        return False

    async def set_prebrew_time(
        self,
        prebrew_on_time: float | None = None,
        prebrew_off_time: float | None = None,
        key: PhysicalKey = PhysicalKey.A,
    ) -> bool:
        """Set prebrew time"""

        if prebrew_on_time is None:
            prebrew_on_time = self.config.prebrew_configuration[key].on_time

        if prebrew_off_time is None:
            prebrew_off_time = self.config.prebrew_configuration[key].off_time

        if await self.cloud_client.configure_pre_brew_infusion_time(
            self.serial_number, prebrew_on_time, prebrew_off_time, key
        ):
            self.config.prebrew_configuration[key].on_time = prebrew_on_time
            self.config.prebrew_configuration[key].off_time = prebrew_off_time
            return True
        return False

    async def set_preinfusion_time(
        self,
        preinfusion_time: float,
        key: PhysicalKey = PhysicalKey.A,
    ) -> bool:
        """Set preinfusion time"""

        if await self.cloud_client.configure_pre_brew_infusion_time(
            self.serial_number, 0, preinfusion_time, key
        ):
            self.config.prebrew_configuration[key].off_time = preinfusion_time
            return True
        return False

    async def set_dose(self, dose: int, key: PhysicalKey) -> bool:
        """Set dose"""

        if await self.cloud_client.set_dose(self.serial_number, key, dose):
            self.config.doses[key] = dose
            return True
        return False

    async def set_dose_tea_water(self, dose: int) -> bool:
        """Set tea dose"""

        if await self.cloud_client.set_dose_hot_water(self.serial_number, dose):
            self.config.dose_hot_water = dose
            return True
        return False

    async def set_scale_target(self, key: PhysicalKey, target: int) -> bool:
        """Set scale target"""
        if not self.model == MachineModel.LINEA_MINI:
            raise ValueError("Scale is only supported on Linea Mini")

        assert self.config.bbw_settings

        if await self.cloud_client.set_scale_target(self.serial_number, key, target):
            self.config.bbw_settings.doses[key] = target
            return True
        return False

    async def set_active_bbw_recipe(self, key: PhysicalKey) -> bool:
        """Set the active scale target"""
        if not self.model == MachineModel.LINEA_MINI:
            raise ValueError("Scale is only supported on Linea Mini")

        assert self.config.bbw_settings

        if await self.cloud_client.set_active_bbw_recipe(self.serial_number, key):
            self.config.bbw_settings.active_dose = key
            return True
        return False

    async def set_bbw_recipe_target(self, key: PhysicalKey, target: int) -> bool:
        """Set the bbw recipe target"""
        if not self.model == MachineModel.LINEA_MINI:
            raise ValueError("Scale is only supported on Linea Mini")

        assert self.config.bbw_settings

        doses = deepcopy(self.config.bbw_settings.doses)
        doses[key] = target

        if await self.cloud_client.set_bbw_recipes(self.serial_number, doses):
            self.config.bbw_settings.doses = doses
            return True
        return False

    async def set_plumbed_in(self, enabled: bool) -> bool:
        """Set plumbed in"""

        if await self.cloud_client.enable_plumbin(self.serial_number, enabled):
            self.config.plumbed_in = enabled
            return True
        return False

    async def start_backflush(self) -> None:
        """Start backflush"""

        await self.cloud_client.start_backflush(self.serial_number)
        self.config.backflush_enabled = True

    # async def set_schedule(self, schedule: LaMarzoccoSchedule) -> bool:
    #     """Set schedule"""

    #     if await self.cloud_client.set_schedule(
    #         self.serial_number, schedule_to_request(schedule)
    #     ):
    #         self.config.auto_on_off_schedule = schedule
    #         return True
    #     return False

    # async def enable_schedule_globally(self, enabled: bool) -> bool:
    #     """Enable schedule globally"""

    #     schedule = deepcopy(self.config.auto_on_off_schedule)
    #     schedule.enabled = enabled
    #     return await self.set_schedule(schedule)

    # async def set_schedule_day(
    #     self,
    #     day: WeekDay,
    #     enabled: bool,
    #     h_on: int,
    #     m_on: int,
    #     h_off: int,
    #     m_off: int,
    # ) -> bool:
    #     """Configure a single day in the schedule"""

    #     day_settings = LaMarzoccoScheduleDay(
    #         enabled=enabled,
    #         h_on=h_on,
    #         h_off=h_off,
    #         m_on=m_on,
    #         m_off=m_off,
    #     )
    #     schedule = deepcopy(self.config.auto_on_off_schedule)
    #     schedule.days[day] = day_settings
    #     return await self.set_schedule(schedule)

    async def set_wake_up_sleep(self, wake_up_sleep_entry: LaMarzoccoWakeUpSleepEntry):
        """Set wake up sleep"""

        if await self.cloud_client.set_wake_up_sleep(
            self.serial_number, wake_up_sleep_entry
        ):
            self.config.wake_up_sleep_entries[wake_up_sleep_entry.entry_id] = (
                wake_up_sleep_entry
            )
            return True
        return False

    async def set_smart_standby(
        self, enabled: bool, minutes: int, mode: SmartStandbyMode
    ) -> bool:
        """Set smart standby"""

        if await self.cloud_client.set_smart_standby(
            serial_number=self.serial_number,
            enabled=enabled,
            minutes=minutes,
            mode=mode,
        ):
            self.config.smart_standby = LaMarzoccoSmartStandby(
                enabled=enabled, mode=mode, minutes=minutes
            )
            return True
        return False

    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""

        return await self.cloud_client.update_firmware(self.serial_number, component)

    async def websocket_connect(
        self, notify_callback: Callable[[], None] | None = None
    ) -> None:
        """Connect to the websocket of the machine."""

        self._notify_callback = notify_callback
        if self._local_client is None:
            raise ClientNotInitialized("Local client not initialized")

        await self._local_client.websocket_connect(
            callback=self.on_websocket_message_received
        )

    async def websocket_disconnect(self) -> None:
        """Disconnect from the websocket of the machine."""

        if self._local_client is None or self._local_client.websocket is None:
            return

        await self._local_client.websocket.close()

    def on_websocket_message_received(self, message: str | bytes) -> None:
        """Websocket message received"""

        self.timestamp_last_websocket_msg = time()
        message = str(message)

        _LOGGER.debug("Received message from websocket, message %s", message)
        notify = False
        try:
            notify = self._parse_websocket_message(message)
        except UnknownWebSocketMessage as exc:
            _LOGGER.warning("Unknown websocket message received")
            _LOGGER.warning("Message: %s", message)
            _LOGGER.debug(exc)
        except TypeError as exc:
            _LOGGER.error("Error when parsing websocket message: %s, %s", message, exc)

        if notify and self._notify_callback:
            self._notify_callback()

    def _parse_websocket_message(self, message: str) -> bool:
        """Handle a message received on the websocket."""

        message = json.loads(message)

        if isinstance(message, dict):
            self._parse_dict_message(message)
            return False

        if isinstance(message, list):
            if "KeepAlive" in message[0]:
                return False
            self._parse_list_message(message)
            return True

        raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")

    def _parse_dict_message(self, message: Any) -> bool:
        """ "Parse websocket message that is a dict."""

        if "MachineConfiguration" in message:
            # got machine configuration
            self._raw_config = json.loads(message["MachineConfiguration"])
            return False

        if "SystemInfo" in message:
            self._system_info = json.loads(message["SystemInfo"])
            return False

        raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")

    def _parse_list_message(self, message: list[dict[str, Any]]) -> bool:
        """Parse websocket message that is a list."""

        property_updated = False
        for msg in message:

            if "SteamBoilerUpdateTemperature" in msg:
                self.config.boilers[BoilerType.STEAM].current_temperature = msg[
                    "SteamBoilerUpdateTemperature"
                ]
                property_updated = True

            elif "CoffeeBoiler1UpdateTemperature" in msg:
                self.config.boilers[BoilerType.COFFEE].current_temperature = msg[
                    "CoffeeBoiler1UpdateTemperature"
                ]
                property_updated = True

            elif "Sleep" in msg:
                self.config.turned_on = False
                property_updated = True

            elif "SteamBoilerEnabled" in msg:
                value = msg["SteamBoilerEnabled"]
                self.config.boilers[BoilerType.STEAM].enabled = value
                property_updated = True

            elif "WakeUp" in msg:
                self.config.turned_on = True
                property_updated = True

            elif "MachineMode" in msg:
                self.config.turned_on = msg["MachineMode"] == "BrewingMode"
                property_updated = True

            elif "MachineStatistics" in msg:
                property_updated = True

            elif "BrewingUpdateGroup1Time" in msg:
                self.config.brew_active_duration = msg["BrewingUpdateGroup1Time"]
                property_updated = True

            elif "BrewingStartedGroup1StopType" in msg:
                self.config.brew_active = True
                property_updated = True

            elif (
                "BrewingStoppedGroup1StopType" in msg
                or "BrewingSnapshotGroup1" in msg
                or "FlushSnapshotGroup1" in msg
            ):
                self.config.brew_active = False
                property_updated = True

            elif "FlushStoppedGroup1Time" in msg:
                self.config.brew_active = False
                self.config.brew_active_duration = msg["FlushStoppedGroup1Time"]
                property_updated = True

            elif "SteamBoilerUpdateSetPoint" in msg:
                self.config.boilers[BoilerType.STEAM].target_temperature = msg[
                    "SteamBoilerUpdateSetPoint"
                ]
                property_updated = True

            elif "CoffeeBoiler1UpdateSetPoint" in msg:
                self.config.boilers[BoilerType.COFFEE].target_temperature = msg[
                    "CoffeeBoiler1UpdateSetPoint"
                ]
                property_updated = True

            elif "BoilersTargetTemperature" in msg:
                target_temps = json.loads(msg["BoilersTargetTemperature"])
                for boiler in BoilerType:
                    self.config.boilers[boiler].target_temperature = target_temps[
                        boiler
                    ]
                property_updated = True

            elif "Boilers" in msg:
                boilers = json.loads(msg["Boilers"])
                self.config.boilers = parse_boilers(boilers)
                property_updated = True

            elif "PreinfusionSettings" in msg:
                settings: dict[str, Any] = {}
                settings["preinfusionSettings"] = json.loads(msg["PreinfusionSettings"])

                self.config.prebrew_mode, self.config.prebrew_configuration = (
                    parse_preinfusion_settings(settings)
                )
                property_updated = True

            elif "SystemInfo" in msg:
                self._system_info = json.loads(msg["SystemInfo"])
                property_updated = True

            elif "TankStatus" in msg:
                self.config.water_contact = msg["TankStatus"] == "full"  # or "empty"
                property_updated = True

            elif "GroupCapabilities" in msg:
                doses = json.loads(msg["GroupCapabilities"])[0]["doses"]
                for dose in doses:
                    key = PhysicalKey[dose["doseIndex"][-1]]
                    self.config.doses[key] = dose["stopTarget"]

        if not property_updated:
            raise UnknownWebSocketMessage(f"Unknown websocket message: {message}")
        return True



"""Base class for all La Marzocco IoT devices."""

import asyncio
import logging
from abc import abstractmethod
from typing import Any

from bleak import BleakError

from pylamarzocco.clients.bluetooth import LaMarzoccoBluetoothClient
from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.clients.local import LaMarzoccoLocalClient
from pylamarzocco.const import FirmwareType
from pylamarzocco.exceptions import (
    AuthFail,
    BluetoothConnectionFailed,
    ClientNotInitialized,
    RequestNotSuccessful,
)
from pylamarzocco.helpers import parse_firmware
from pylamarzocco.models import (
    LaMarzoccoDeviceConfig,
    LaMarzoccoFirmware,
    LaMarzoccoStatistics,
)

_LOGGER = logging.getLogger(__name__)


class LaMarzoccoBaseDevice:
    """Base class for all La Marzocco devices"""

    _cloud_client: LaMarzoccoCloudClient | None

    def __init__(
        self,
        model: str,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
        bluetooth_client: LaMarzoccoBluetoothClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoMachine instance"""

        self.model: str = model
        self.serial_number: str = serial_number
        self.name: str = name
        self.firmware: dict[FirmwareType, LaMarzoccoFirmware] = {}
        self.statistics: LaMarzoccoStatistics | None = None
        self.config: LaMarzoccoDeviceConfig = LaMarzoccoDeviceConfig(
            turned_on=False,
            doses={},
        )

        self._raw_config: dict[str, Any] | None = None
        self._cloud_client = cloud_client
        self._bluetooth_client = bluetooth_client
        self._local_client = local_client
        self._update_lock = asyncio.Lock()
        self._local_connection_down = False

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        self.firmware = parse_firmware(raw_config["firmwareVersions"], self.firmware)

    @property
    def cloud_client(self) -> LaMarzoccoCloudClient:
        """Ensure that the cloud client is initialized."""

        if self._cloud_client is None:
            raise ClientNotInitialized("Cloud client not initialized")
        return self._cloud_client

    @abstractmethod
    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

    @property
    def full_model_name(self) -> str:
        """Return the full model name of the device."""
        return self.model

    async def get_config(
        self,
        local_api_retry_delay: int = 3,
    ) -> None:
        """Update the machine status."""

        raw_config: dict[str, Any] = {}

        async with self._update_lock:
            # first, try to update locally
            if self._local_client is not None:
                try:
                    raw_config = await self._local_client.get_config()
                except AuthFail as exc:
                    _LOGGER.debug(
                        "Got 403 from local API, sending token request to cloud"
                    )
                    # for some machines (GS3) we need to send a token request before local API works
                    if self._cloud_client is not None:
                        await self._cloud_client.token_command(self.serial_number)
                        await asyncio.sleep(local_api_retry_delay)
                        raw_config = await self._local_client.get_config()
                    else:
                        raise exc
                except RequestNotSuccessful as exc:
                    msg = (
                        "Could not connect to local API although initialized, "
                        "falling back to cloud."
                    )
                    if not self._local_connection_down:
                        _LOGGER.warning(msg)
                        self._local_connection_down = True
                    _LOGGER.debug(msg)
                    _LOGGER.debug(exc)
                    if self._cloud_client is None:
                        raise exc
                else:
                    if self._local_connection_down:
                        _LOGGER.warning("Local API connection restored.")
                        self._local_connection_down = False

            # if local update failed, try to update from cloud
            if self._cloud_client is not None and not raw_config:
                raw_config = await self._cloud_client.get_config(self.serial_number)

            self.parse_config(raw_config)

    async def get_statistics(self) -> None:
        """Update the statistics"""

        raw_statistics = await self.cloud_client.get_statistics(self.serial_number)
        self.parse_statistics(raw_statistics)

    async def get_firmware(self) -> None:
        """Update the firmware"""

        self.firmware = await self.cloud_client.get_firmware(self.serial_number)

    @abstractmethod
    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""

    async def _bluetooth_command_with_cloud_fallback(
        self,
        command: str,
        **kwargs,
    ) -> bool:
        """Send a command to the machine via Bluetooth, falling back to cloud if necessary."""

        # First, try with bluetooth
        if self._bluetooth_client is not None:
            func = getattr(self._bluetooth_client, command)
            try:
                _LOGGER.debug(
                    "Sending command %s over bluetooth with params %s",
                    command,
                    str(kwargs),
                )
                await func(**kwargs)
            except (BleakError, BluetoothConnectionFailed) as exc:
                msg = "Could not send command to bluetooth device, even though initalized."

                if self._cloud_client is None:
                    _LOGGER.error(
                        "%s Cloud client not initialized, cannot fallback. Full error %s",
                        msg,
                        exc,
                    )
                    return False

                _LOGGER.warning("%s Falling back to cloud", msg)
                _LOGGER.debug("Full error: %s", exc)
            else:
                return True

        # no bluetooth or failed, try with cloud
        if self._cloud_client is not None:
            _LOGGER.debug(
                "Sending command %s over cloud with params %s",
                command,
                str(kwargs),
            )
            func = getattr(self._cloud_client, command)
            kwargs["serial_number"] = self.serial_number
            if await func(**kwargs):
                return True
        return False

    def __str__(self) -> str:
        config = dict(vars(self.config).items())
        return str(config)



"""Test the LaMarzoccoGrinder class."""

from syrupy import SnapshotAssertion

from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.const import GrinderModel
from pylamarzocco.devices.grinder import LaMarzoccoGrinder

from . import GRINDER_SERIAL


async def test_create(
    cloud_client: LaMarzoccoCloudClient,
    snapshot: SnapshotAssertion,
) -> None:
    """Test creation of a cloud client."""

    machine = await LaMarzoccoGrinder.create(
        model=GrinderModel("Pico"),
        serial_number=GRINDER_SERIAL,
        name="MyMachine",
        cloud_client=cloud_client,
    )
    assert machine == snapshot



# """"Test the Bluetooth client."""

# from collections.abc import Generator
# from unittest.mock import AsyncMock

# from pylamarzocco.clients.bluetooth import LaMarzoccoBluetoothClient
# from pylamarzocco.const import (
#     AUTH_CHARACTERISTIC,
#     SETTINGS_CHARACTERISTIC,
#     BoilerType,
# )


# async def test_bluetooth_set_power(
#     bluetooth_client: LaMarzoccoBluetoothClient,
#     mock_bleak: Generator[AsyncMock, None, None],
# ):
#     """Test setting the power."""
#     await bluetooth_client.set_power(True)

#     mock_bleak.write_gatt_char.assert_any_call(  # type: ignore[attr-defined]
#         char_specifier=AUTH_CHARACTERISTIC,
#         data=b"dXNlcm5hbWU6c2VyaWFs@dG9rZW4=",
#         response=True,
#     )

#     mock_bleak.write_gatt_char.assert_called_with(  # type: ignore[attr-defined]
#         char_specifier=SETTINGS_CHARACTERISTIC,
#         data=b'{"name":"MachineChangeMode","parameter":{"mode":"BrewingMode"}}\x00',
#         response=True,
#     )


# async def test_bluetooth_set_steam(
#     bluetooth_client: LaMarzoccoBluetoothClient,
#     mock_bleak: Generator[AsyncMock, None, None],
# ):
#     """Test setting the steam."""
#     await bluetooth_client.set_steam(True)

#     mock_bleak.write_gatt_char.assert_any_call(  # type: ignore[attr-defined]
#         char_specifier=AUTH_CHARACTERISTIC,
#         data=b"dXNlcm5hbWU6c2VyaWFs@dG9rZW4=",
#         response=True,
#     )

#     mock_bleak.write_gatt_char.assert_called_with(  # type: ignore[attr-defined]
#         char_specifier=SETTINGS_CHARACTERISTIC,
#         data=b'{"name":"SettingBoilerEnable","parameter":{"identifier":"SteamBoiler","state":true}}\x00',
#         response=True,
#     )


# async def test_bluetooth_set_temperature(
#     bluetooth_client: LaMarzoccoBluetoothClient,
#     mock_bleak: Generator[AsyncMock, None, None],
# ):
#     """Test setting the temp."""
#     await bluetooth_client.set_temp(BoilerType.STEAM, 131)

#     mock_bleak.write_gatt_char.assert_any_call(  # type: ignore[attr-defined]
#         char_specifier=AUTH_CHARACTERISTIC,
#         data=b"dXNlcm5hbWU6c2VyaWFs@dG9rZW4=",
#         response=True,
#     )

#     mock_bleak.write_gatt_char.assert_called_with(  # type: ignore[attr-defined]
#         char_specifier=SETTINGS_CHARACTERISTIC,
#         data=b'{"name":"SettingBoilerTarget","parameter":{"identifier":"SteamBoiler","value":131}}\x00',
#         response=True,
#     )



"""Test the LaMarzoccoMachine class."""

# pylint: disable=W0212
from dataclasses import asdict
from http import HTTPMethod
from unittest.mock import AsyncMock, patch
from aiohttp import ClientTimeout
import pytest

from aioresponses import aioresponses
from bleak import BleakError
from syrupy import SnapshotAssertion

from pylamarzocco.clients.bluetooth import LaMarzoccoBluetoothClient
from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.clients.local import LaMarzoccoLocalClient
from pylamarzocco.const import (
    BoilerType,
    PhysicalKey,
    MachineModel,
    GW_MACHINE_BASE_URL,
)
from pylamarzocco.devices.machine import LaMarzoccoMachine
from pylamarzocco.models import LaMarzoccoBrewByWeightSettings

from . import init_machine
from .conftest import load_fixture

CLIENT_TIMEOUT = ClientTimeout(
    total=5, connect=None, sock_read=None, sock_connect=None, ceil_threshold=5
)


async def test_create(
    cloud_client: LaMarzoccoCloudClient,
    snapshot: SnapshotAssertion,
) -> None:
    """Test creation of a cloud client."""

    machine = await init_machine(cloud_client)
    assert asdict(machine.config) == snapshot(name="config")
    assert machine.firmware == snapshot(name="firmware")
    assert machine.statistics == snapshot(name="statistics")


async def test_mini(
    cloud_client: LaMarzoccoCloudClient,
    mock_aioresponse: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test creation of a cloud client."""
    serial = "LM01234"
    mock_aioresponse.get(
        url=f"{GW_MACHINE_BASE_URL}/{serial}/configuration",
        status=200,
        payload=load_fixture("machine", "config-mini.json"),
    )
    mock_aioresponse.get(
        url=f"{GW_MACHINE_BASE_URL}/{serial}/statistics/counters",
        status=200,
        payload=load_fixture("machine", "counters.json"),
    )
    mock_aioresponse.get(
        url=f"{GW_MACHINE_BASE_URL}/{serial}/firmware/",
        status=200,
        payload=load_fixture("machine", "firmware.json"),
    )
    machine = await LaMarzoccoMachine.create(
        model=MachineModel.LINEA_MINI,
        serial_number=serial,
        name="MyMachine",
        cloud_client=cloud_client,
    )

    assert asdict(machine.config) == snapshot


async def test_local_client(
    machine: LaMarzoccoMachine,
    local_machine_client: LaMarzoccoLocalClient,
    mock_aioresponse: aioresponses,
) -> None:
    """Ensure that the local client delivers same result"""
    # load config
    mock_aioresponse.get(
        url="http://192.168.1.42:8081/api/v1/config",
        status=200,
        payload=load_fixture("machine", "config.json")["data"],
    )

    machine_local = await init_machine(local_client=local_machine_client)

    assert machine_local
    assert str(machine.config) == str(machine_local.config)


async def test_set_temp(machine: LaMarzoccoMachine) -> None:
    """Test setting boiler temperature."""

    result = await machine.set_temp(
        BoilerType.STEAM,
        120,
    )
    assert result is True
    assert machine.config.boilers[BoilerType.STEAM].target_temperature == 120


# async def test_set_schedule(
#     cloud_client: LaMarzoccoCloudClient,
# ) -> None:
#     """Test setting prebrew infusion."""
#     machine = await init_machine(cloud_client)

#     with patch("asyncio.sleep", new_callable=AsyncMock):
#         result = await machine.set_schedule_day(
#             day=WeekDay.MONDAY,
#             enabled=True,
#             h_on=3,
#             m_on=0,
#             h_off=24,
#             m_off=0,
#         )
#     assert result is True


async def test_set_power(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting the power."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    with patch(
        "pylamarzocco.clients.bluetooth.LaMarzoccoBluetoothClient.set_power",
        new=AsyncMock(),
    ) as mock_set_power:
        mock_set_power.side_effect = BleakError("Failed to write")
        assert await machine.set_power(True)

        mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
            method=HTTPMethod.POST,
            url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/status",
            json={"status": "BrewingMode"},
            headers={"Authorization": "Bearer 123"},
            timeout=CLIENT_TIMEOUT,
            allow_redirects=True,
            data=None,
        )


async def test_set_steam(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting the steam."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    with patch(
        "pylamarzocco.clients.bluetooth.LaMarzoccoBluetoothClient.set_steam",
        new=AsyncMock(),
    ) as mock_set_steam:
        mock_set_steam.side_effect = BleakError("Failed to write")

        assert await machine.set_steam(True)

        mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
            method=HTTPMethod.POST,
            url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/enable-boiler",
            json={"identifier": "SteamBoiler", "state": True},
            headers={"Authorization": "Bearer 123"},
            timeout=CLIENT_TIMEOUT,
            allow_redirects=True,
            data=None,
        )


async def test_set_temperature(
    cloud_client: LaMarzoccoCloudClient,
    bluetooth_client: LaMarzoccoBluetoothClient,
    mock_aioresponse: aioresponses,
):
    """Test setting temperature."""
    machine = await init_machine(cloud_client, bluetooth_client=bluetooth_client)

    with patch(
        "pylamarzocco.clients.bluetooth.LaMarzoccoBluetoothClient.set_temp",
        new=AsyncMock(),
    ) as mock_set_temp:
        mock_set_temp.side_effect = BleakError("Failed to write")

        assert await machine.set_temp(BoilerType.STEAM, 131)

        mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
            method=HTTPMethod.POST,
            url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/target-boiler",
            json={"identifier": "SteamBoiler", "value": 131},
            headers={"Authorization": "Bearer 123"},
            timeout=CLIENT_TIMEOUT,
            allow_redirects=True,
            data=None,
        )


async def test_set_prebrew_time(
    machine: LaMarzoccoMachine,
    mock_aioresponse: aioresponses,
):
    """Test setting prebrew time."""

    assert await machine.set_prebrew_time(1.0, 3.5)

    mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/setting-preinfusion",
        json={
            "button": "DoseA",
            "group": "Group1",
            "holdTimeMs": 3500,
            "wetTimeMs": 1000,
        },
        headers={"Authorization": "Bearer 123"},
        timeout=CLIENT_TIMEOUT,
        allow_redirects=True,
        data=None,
    )

    assert machine.config.prebrew_configuration[PhysicalKey.A].on_time == 1.0
    assert machine.config.prebrew_configuration[PhysicalKey.A].off_time == 3.5


async def test_set_preinfusion_time(
    machine: LaMarzoccoMachine,
    mock_aioresponse: aioresponses,
):
    """Test setting prebrew time."""
    assert await machine.set_preinfusion_time(4.5)
    mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/setting-preinfusion",
        json={"button": "DoseA", "group": "Group1", "holdTimeMs": 4500, "wetTimeMs": 0},
        headers={"Authorization": "Bearer 123"},
        timeout=CLIENT_TIMEOUT,
        allow_redirects=True,
        data=None,
    )

    assert machine.config.prebrew_configuration[PhysicalKey.A].off_time == 4.5


async def test_set_scale_target(
    machine: LaMarzoccoMachine,
    mock_aioresponse: aioresponses,
):
    """Test setting scale target."""
    # fails with not Linea Mini
    with pytest.raises(ValueError):
        await machine.set_scale_target(PhysicalKey.B, 42)

    # set to Linea Mini
    machine.model = MachineModel.LINEA_MINI
    machine.config.bbw_settings = LaMarzoccoBrewByWeightSettings(
        doses={}, active_dose=PhysicalKey.A
    )

    assert await machine.set_scale_target(PhysicalKey.B, 42)
    mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/scale/target-dose",
        json={
            "group": "Group1",
            "dose_index": "DoseB",
            "dose_type": "MassType",
            "value": 42,
        },
        headers={"Authorization": "Bearer 123"},
        timeout=CLIENT_TIMEOUT,
        allow_redirects=True,
        data=None,
    )

    assert machine.config.bbw_settings.doses[PhysicalKey.B] == 42


async def test_set_bbw_recipe(
    machine: LaMarzoccoMachine,
    mock_aioresponse: aioresponses,
):
    """Test setting scale target."""
    # fails with not Linea Mini
    with pytest.raises(ValueError):
        await machine.set_bbw_recipe_target(PhysicalKey.B, 42)

    # set to Linea Mini
    machine.model = MachineModel.LINEA_MINI
    machine.config.bbw_settings = LaMarzoccoBrewByWeightSettings(
        doses={PhysicalKey.A: 12, PhysicalKey.B: 34}, active_dose=PhysicalKey.A
    )

    assert await machine.set_bbw_recipe_target(PhysicalKey.B, 42)
    mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
        method=HTTPMethod.PUT,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/recipes/",
        json={
            "recipeId": "Recipe1",
            "doseMode": "Mass",
            "recipeDoses": [
                {"id": "A", "target": 12},
                {"id": "B", "target": 42},
            ],
        },
        headers={"Authorization": "Bearer 123"},
        timeout=CLIENT_TIMEOUT,
        allow_redirects=True,
    )

    assert machine.config.bbw_settings.doses[PhysicalKey.B] == 42


async def test_set_active_bbw_recipe(
    machine: LaMarzoccoMachine,
    mock_aioresponse: aioresponses,
):
    """Test active bbw recipe."""
    # fails with not Linea Mini
    with pytest.raises(ValueError):
        await machine.set_active_bbw_recipe(PhysicalKey.B)

    # set to Linea Mini
    machine.model = MachineModel.LINEA_MINI
    machine.config.bbw_settings = LaMarzoccoBrewByWeightSettings(
        doses={PhysicalKey.A: 12, PhysicalKey.B: 34}, active_dose=PhysicalKey.A
    )

    assert await machine.set_active_bbw_recipe(PhysicalKey.B)
    mock_aioresponse.assert_called_with(  # type: ignore[attr-defined]
        method=HTTPMethod.POST,
        url="https://gw-lmz.lamarzocco.io/v1/home/machines/GS01234/recipes/active-recipe",
        json={
            "group": "Group1",
            "doseIndex": "DoseA",
            "recipeId": "Recipe1",
            "recipeDose": "B",
        },
        headers={"Authorization": "Bearer 123"},
        timeout=CLIENT_TIMEOUT,
        allow_redirects=True,
        data=None,
    )

    assert machine.config.bbw_settings.active_dose == PhysicalKey.B



"""Bluetooth class for La Marzocco machines."""

from __future__ import annotations

import asyncio
import base64
import json
import logging
from typing import Any

from bleak import (
    BaseBleakScanner,
    BleakClient,
    BleakError,
    BleakScanner,
    BLEDevice,
)

from pylamarzocco.const import (
    AUTH_CHARACTERISTIC,
    BT_MODEL_PREFIXES,
    SETTINGS_CHARACTERISTIC,
    BoilerType,
)
from pylamarzocco.exceptions import BluetoothConnectionFailed

_logger = logging.getLogger(__name__)


class LaMarzoccoBluetoothClient:
    """Class to interact with machine via Bluetooth."""

    def __init__(
        self,
        username: str,
        serial_number: str,
        token: str,
        address_or_ble_device: BLEDevice | str,
    ) -> None:
        """Initializes a new LaMarzoccoBluetoothClient instance."""
        self._username = username
        self._serial_number = serial_number
        self._token = token
        self._address = (
            address_or_ble_device.address
            if isinstance(address_or_ble_device, BLEDevice)
            else address_or_ble_device
        )
        self._address_or_ble_device = address_or_ble_device

    @staticmethod
    async def discover_devices(
        scanner: BaseBleakScanner | BleakScanner | None = None,
    ) -> list[BLEDevice]:
        """Find machines based on model name."""
        ble_devices: list[BLEDevice] = []

        if scanner is None:
            scanner = BleakScanner()
        assert hasattr(scanner, "discover")
        devices: list[BLEDevice] = await scanner.discover()
        for device in devices:
            if device.name and device.name.startswith(BT_MODEL_PREFIXES):
                ble_devices.append(device)

        return ble_devices

    @property
    def address(self) -> str:
        """Return the BT MAC address of the machine."""

        return self._address

    async def set_power(self, enabled: bool) -> None:
        """Power on the machine."""

        mode = "BrewingMode" if enabled else "StandBy"
        data = {
            "name": "MachineChangeMode",
            "parameter": {
                "mode": mode,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def set_steam(self, enabled: bool) -> None:
        """Power cycle steam."""

        data = {
            "name": "SettingBoilerEnable",
            "parameter": {
                "identifier": "SteamBoiler",
                "state": enabled,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def set_temp(self, boiler: BoilerType, temperature: float) -> None:
        """Set boiler temperature (in Celsius)"""

        data = {
            "name": "SettingBoilerTarget",
            "parameter": {
                "identifier": boiler.value,
                "value": temperature,
            },
        }
        await self._write_bluetooth_json_message(data)

    async def _write_bluetooth_message(
        self, characteristic: str, message: bytes | str
    ) -> None:
        """Connect to machine and write a message."""

        # check if message is already bytes string
        if not isinstance(message, bytes):
            message = bytes(message, "utf-8")

        # append trailing zeros to message
        message += b"\x00"

        async with BleakClient(self._address_or_ble_device) as client:

            async def service_info() -> None:
                for service in client.services:
                    _logger.debug("[Service] %s", service)

                    for char in service.characteristics:
                        if "read" in char.properties:
                            try:
                                value = await client.read_gatt_char(char.uuid)
                                extra = f", Value: {value}"
                            except (
                                Exception  # pylint: disable=broad-exception-caught
                            ) as e:
                                extra = f", Error: {e}"
                        else:
                            extra = ""

                        if "write-without-response" in char.properties:
                            extra += f", Max write w/o rsp size: {char.max_write_without_response_size}"

                        _logger.debug(
                            "  [Characteristic] %s (%s)%s",
                            char,
                            ",".join(char.properties),
                            extra,
                        )

                        for descriptor in char.descriptors:
                            try:
                                value = await client.read_gatt_descriptor(
                                    descriptor.handle
                                )
                                _logger.debug(
                                    "    [Descriptor] %s, Value: %r", descriptor, value
                                )
                            except (
                                Exception  # pylint: disable=broad-exception-caught
                            ) as e:
                                _logger.error(
                                    "    [Descriptor] %s, Error: %s", descriptor, e
                                )

            async def authenticate() -> None:
                """Build authentication string and send it to the machine."""

                user = self._username + ":" + self._serial_number
                user_bytes = user.encode("utf-8")
                token = self._token.encode("utf-8")
                auth_string = (
                    base64.b64encode(user_bytes) + b"@" + base64.b64encode(token)
                )

                auth_characteristic = client.services.get_characteristic(
                    AUTH_CHARACTERISTIC
                )

                if auth_characteristic is None:
                    raise BluetoothConnectionFailed(
                        f"Could not find auth characteristic {AUTH_CHARACTERISTIC} on machine."
                    )
                _logger.debug("Auth characteristic: %s", auth_characteristic.handle)
                try:
                    await client.write_gatt_char(
                        char_specifier=auth_characteristic,
                        data=auth_string,
                        response=True,
                    )
                except (BleakError, TimeoutError) as e:
                    raise BluetoothConnectionFailed(
                        f"Failed to connect to machine with Bluetooth: {e}"
                    ) from e

            await service_info()
            await authenticate()
            await asyncio.sleep(0.1)
            _logger.debug(
                "Sending bluetooth message: %s to %s", message, characteristic
            )

            settings_characteristic = client.services.get_characteristic(
                SETTINGS_CHARACTERISTIC
            )

            if settings_characteristic is None:
                raise BluetoothConnectionFailed(
                    f"Could not find settings characteristic {SETTINGS_CHARACTERISTIC} on machine."
                )
            _logger.debug("Settings characteristic: %s", settings_characteristic.handle)

            await client.write_gatt_char(
                char_specifier=settings_characteristic,
                data=message,
                response=True,
            )

    async def _write_bluetooth_json_message(
        self,
        data: dict[str, Any],
        characteristic: str = SETTINGS_CHARACTERISTIC,
    ) -> None:
        """Write a json message to the machine."""

        await self._write_bluetooth_message(
            characteristic=characteristic,
            message=json.dumps(data, separators=(",", ":")),
        )



""""Test websocket messages parsing."""

from dataclasses import asdict

from syrupy import SnapshotAssertion

from pylamarzocco.devices.machine import LaMarzoccoMachine

from pylamarzocco.const import PhysicalKey


async def test_websocket_message(
    machine: LaMarzoccoMachine,
    snapshot: SnapshotAssertion,
):
    """Test parsing of websocket messages."""

    message = r'[{"Boilers":"[{\"id\":\"SteamBoiler\",\"isEnabled\":true,\"target\":131,\"current\":113},{\"id\":\"CoffeeBoiler1\",\"isEnabled\":true,\"target\":94,\"current\":81}]"}]'
    machine.on_websocket_message_received(message)
    assert asdict(machine.config) == snapshot

    message = r'[{"BoilersTargetTemperature":"{\"SteamBoiler\":131,\"CoffeeBoiler1\":94}"},{"Boilers":"[{\"id\":\"SteamBoiler\",\"isEnabled\":true,\"target\":131,\"current\":50},{\"id\":\"CoffeeBoiler1\",\"isEnabled\":true,\"target\":94,\"current\":36}]"}]'
    machine.on_websocket_message_received(message)
    assert asdict(machine.config) == snapshot


async def test_group_capabilities_websocket_message(
    machine: LaMarzoccoMachine,
):
    """Test parsing of group capabilities websocket message."""
    msg = '[{"GroupCapabilities": "[{\\"capabilities\\":{\\"groupType\\":\\"AV_Group\\",\\"groupNumber\\":\\"Group1\\",\\"boilerId\\":\\"CoffeeBoiler1\\",\\"hasScale\\":false,\\"hasFlowmeter\\":true,\\"numberOfDoses\\":4},\\"doses\\":[{\\"groupNumber\\":\\"Group1\\",\\"doseIndex\\":\\"DoseA\\",\\"doseType\\":\\"PulsesType\\",\\"stopTarget\\":126},{\\"groupNumber\\":\\"Group1\\",\\"doseIndex\\":\\"DoseB\\",\\"doseType\\":\\"PulsesType\\",\\"stopTarget\\":130},{\\"groupNumber\\":\\"Group1\\",\\"doseIndex\\":\\"DoseC\\",\\"doseType\\":\\"PulsesType\\",\\"stopTarget\\":140},{\\"groupNumber\\":\\"Group1\\",\\"doseIndex\\":\\"DoseD\\",\\"doseType\\":\\"PulsesType\\",\\"stopTarget\\":77}],\\"doseMode\\":{\\"groupNumber\\":\\"Group1\\",\\"brewingType\\":\\"PulsesType\\"}}]"}]'
    machine.on_websocket_message_received(msg)
    assert machine.config.doses[PhysicalKey(1)] == 126
    assert machine.config.doses[PhysicalKey(2)] == 130
    assert machine.config.doses[PhysicalKey(3)] == 140
    assert machine.config.doses[PhysicalKey(4)] == 77


async def test_flush_snapshot_message(machine: LaMarzoccoMachine):
    """Test the flush snapshot message"""
    machine.config.brew_active = True
    msg = r'[{"FlushSnapshotGroup1": "{\"groupConfiguration\":{\"groupNumber\":\"Group1\",\"capabilities\":{\"groupType\":\"AV_Group\",\"boilerId\":\"CoffeeBoiler1\",\"boilerTemperature\":97,\"boilerTemperatureAtStart\":97,\"hasScale\":false,\"hasFlowmeter\":true},\"dose\":{\"doseIndex\":\"DoseA\",\"stopTarget\":0,\"doseType\":\"PulsesType\"},\"doseMode\":{\"brewingType\":\"PulsesType\"}},\"flushInfo\":{\"doseIndex\":\"DoseA\",\"stopReason\":\"Manual\",\"time\":4.521,\"stopType\":\"Volumetric\",\"volume\":0}}"},{ "FlushStoppedGroup1DoseIndex": "DoseA" },{ "FlushStoppedGroup1Time": 4.521 },{ "FlushStoppedGroup1Volume": 0 }]'
    machine.on_websocket_message_received(msg)
    assert not machine.config.brew_active
    assert machine.config.brew_active_duration == 4.521



"""Constants for La Marzocco Cloud."""

from enum import IntEnum, StrEnum
from typing import Final


class MachineModel(StrEnum):
    """La Marzocco Coffee Machine Models."""

    GS3_AV = "GS3 AV"
    GS3_MP = "GS3 MP"
    LINEA_MINI = "Linea Mini"
    LINEA_MICRA = "Micra"


class GrinderModel(StrEnum):
    """La Marzocco Grinder Models."""

    PICO = "Pico"


class FirmwareType(StrEnum):
    """La Marzocco updateable firmware components."""

    MACHINE = "machine"
    GATEWAY = "gateway"


class BoilerType(StrEnum):
    """La Marzocco Coffee Machine Boilers."""

    COFFEE = "CoffeeBoiler1"
    STEAM = "SteamBoiler"


class PrebrewMode(StrEnum):
    """Enum for prebrew infusion mode"""

    DISABLED = "Disabled"
    PREBREW = "Enabled"
    PREINFUSION = "TypeB"


class PhysicalKey(IntEnum):
    """Describe the physical keys of the La Marzocco devices."""

    A = 1
    B = 2
    C = 3
    D = 4


class SteamLevel(IntEnum):
    """Enum for steam level."""

    LEVEL_1 = 126
    LEVEL_2 = 128
    LEVEL_3 = 131


class WeekDay(StrEnum):
    """Enum for days of the week."""

    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"
    SATURDAY = "saturday"
    SUNDAY = "sunday"


class SmartStandbyMode(StrEnum):
    """Enum for smart standby mode."""

    LAST_BREWING = "LastBrewing"
    POWER_ON = "PowerOn"


KEYS_PER_MODEL: Final = {
    MachineModel.LINEA_MICRA: 0,
    MachineModel.LINEA_MINI: 0,
    MachineModel.GS3_AV: 4,
    MachineModel.GS3_MP: 0,
}


# Base URL for La Marzocco Cloud
CMS_URL: Final = "https://cms.lamarzocco.io"
TOKEN_URL: Final = f"{CMS_URL}/oauth/v2/token"
LOGOUT_URL: Final = f"{CMS_URL}/oauth/v2/logout"
CUSTOMER_URL: Final = f"{CMS_URL}/api/customer"
GW_BASE_URL: Final = "https://gw-lmz.lamarzocco.io/v1/home"
GW_MACHINE_BASE_URL: Final = f"{GW_BASE_URL}/machines"
GW_AWS_PROXY_BASE_URL: Final = f"{GW_BASE_URL}/aws-proxy"

DEFAULT_CLIENT_ID: Final = "7_1xwei9rtkuckso44ks4o8s0c0oc4swowo00wgw0ogsok84kosg"
DEFAULT_CLIENT_SECRET: Final = "2mgjqpikbfuok8g4s44oo4gsw0ks44okk4kc4kkkko0c8soc8s"
DEFAULT_PORT: Final = 8081

WEBSOCKET_RETRY_DELAY: Final = 20


# bluetooth
BT_MODEL_PREFIXES: Final = ("MICRA", "MINI", "GS3")
SETTINGS_CHARACTERISTIC: Final = "050b7847-e12b-09a8-b04b-8e0922a9abab"
AUTH_CHARACTERISTIC: Final = "090b7847-e12b-09a8-b04b-8e0922a9abab"



"""La Marzocco grinder module."""

from __future__ import annotations

from typing import Any

from pylamarzocco.clients.cloud import LaMarzoccoCloudClient
from pylamarzocco.clients.local import LaMarzoccoLocalClient
from pylamarzocco.const import FirmwareType, GrinderModel, PhysicalKey
from pylamarzocco.devices.base import LaMarzoccoBaseDevice
from pylamarzocco.models import LaMarzoccoGrinderConfig


class LaMarzoccoGrinder(LaMarzoccoBaseDevice):
    """Class for La Marzocco grinder"""

    def __init__(
        self,
        model: GrinderModel,
        serial_number: str,
        name: str,
        cloud_client: LaMarzoccoCloudClient | None = None,
        local_client: LaMarzoccoLocalClient | None = None,
    ) -> None:
        """Initializes a new LaMarzoccoGrinder instance."""

        super().__init__(model, serial_number, name, cloud_client, local_client)
        self.config: LaMarzoccoGrinderConfig = LaMarzoccoGrinderConfig(
            turned_on=False,
            led_enabled=False,
            bell_opened=False,
            stand_by_time=0,
            doses={},
        )

    @classmethod
    async def create(cls, *args, **kwargs) -> LaMarzoccoGrinder:
        """Create a new LaMarzoccoGrinder instance."""

        self = LaMarzoccoGrinder(*args, **kwargs)
        await self.get_config()
        return self

    def parse_config(self, raw_config: dict[str, Any]) -> None:
        """Parse the config object."""

        super().parse_config(raw_config)
        self.config.turned_on = raw_config["machineMode"] == "StandBy"
        self.config.led_enabled = raw_config["baristaLed"]
        self.config.bell_opened = raw_config["bellOpened"]
        self.config.stand_by_time = raw_config["standByTime"]
        for dose in raw_config["doses"]:
            self.config.doses[PhysicalKey(ord(dose["doseIndex"][-1]) - 64)] = dose[
                "target"
            ]

    def parse_statistics(self, raw_statistics: list[dict[str, Any]]) -> None:
        """Parse the statistics object."""

        raise NotImplementedError

    async def update_firmware(self, component: FirmwareType) -> bool:
        """Update firmware"""
        raise NotImplementedError
